package convo;

import java.nio.FloatBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;
import org.lwjgl.opengl.Display;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.IOUtil;
import pa.util.IOUtil.TextureData;
import pa.util.SizeOf;
import visualize.gl.GLUtil;
import visualize.gl.Texture;

public class Blur extends BlurBasis
{
    public Blur() 
    {
        super(1024, 768);
    }

    private CLMem m_clTextureSrc;
    private CLMem m_clTextureDst;
    private CLMem m_filterMask = null;
    private Texture m_glTextureSrc;
    private Texture m_glTextureDst;
    private CLContext m_context;
    private CLKernel m_kernel;
    private CLCommandQueue m_queue;
    private CLProgram m_clProgram;

    private PointerBuffer m_globalWorkSize = BufferUtils.createPointerBuffer(2);
    private PlatformDevicePair pair;
    
    private FloatBuffer m_readBuffer;
    
    private static class Settings
    {
        public static int m_filterSize = 11; //start with an odd number so it stays odd
        public static int MAX_FILTER_SIZE = 55;
        public static double m_sigma = 5;
        public static double m_sigmaDelta = 0.5;
        public static double m_filterSizeDelta = 2;
        
        public static void print()
        {
            System.out.println(String.format("FilterSize=%d, Sigma=%f", Settings.m_filterSize, Settings.m_sigma));
        }
    }

    @Override
    public void render() 
    {
        OpenCL.clEnqueueNDRangeKernel(m_queue, m_kernel, 2, null, m_globalWorkSize, null, null, null);
        OpenCL.clEnqueueReadBuffer(m_queue, m_clTextureDst, 1, 0, m_readBuffer, null, null);
        
        m_glTextureDst.loadFloatData(m_readBuffer);
        
        GLUtil.transformScreenQuad(0, 0, getWidth(), getHeight()).bind();
        GLUtil.drawTexture(1);
        GLUtil.transformScreenQuad(0, 0, getWidth() / 3, getHeight() / 3).bind();
        GLUtil.drawTexture(0);
    }
    
    public float[] getGaussianBlur(int size, double sigma)
    {
        float data[] = new float[size * size];
        int halfSize = size / 2;
        for(int i = 0; i < size; ++i)
        {
            for(int j = 0; j < size; ++j)
            {
                double x = i - halfSize;
                double y = j - halfSize;
                data[i * size + j] = (float)(1.0 / (Math.PI * 2 * sigma * sigma) * Math.exp(-(x * x + y * y) / (2 * sigma * sigma)));
            }
        }
        return data;
    }

    @Override
    public void init() 
    {
                
        TextureData td = IOUtil.readTextureData("textures/brands.jpg");
        FloatBuffer rgbadata = Texture.createRGBAFromX(td);
        
        m_glTextureSrc = Texture.createRGBA2DTexture(td.w, td.h, 0, rgbadata);
        
        m_glTextureDst = Texture.createRGBA2DTexture(td.w, td.h, 1, null);

        GLUtil.transformScreenQuad(0, 0, getWidth(), getHeight()).bind();
        
        CLUtil.createCL();
        
        m_readBuffer = BufferUtils.createFloatBuffer(td.h * td.w * 4);
        
        pair = CLUtil.choosePlatformAndDevice();
        
        m_context = OpenCL.clCreateContext(pair.platform, pair.device, null, Display.getDrawable());
        
        m_queue = OpenCL.clCreateCommandQueue(m_context, pair.device, 0);
        
        m_clProgram = OpenCL.clCreateProgramWithSource(m_context, IOUtil.readFileContent("kernel/convolution.cl"));
        
        OpenCL.clBuildProgram(m_clProgram, pair.device, "", null);
        
        m_kernel = OpenCL.clCreateKernel(m_clProgram, "blur");

        m_clTextureSrc = OpenCL.clCreateBuffer(m_context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, rgbadata);
        m_clTextureDst = OpenCL.clCreateBuffer(m_context, CL10.CL_MEM_WRITE_ONLY, td.h * td.w * 4 * SizeOf.FLOAT);
        
        OpenCL.clSetKernelArg(m_kernel, 0, m_clTextureSrc);
        OpenCL.clSetKernelArg(m_kernel, 1, m_clTextureDst);
        OpenCL.clSetKernelArg(m_kernel, 4, td.w * td.h);
        
        setMask();

        m_globalWorkSize.put(new long[] { td.w, td.h });
        m_globalWorkSize.position(0);
    }

    @Override
    public void close() 
    {
        OpenCL.clReleaseKernel(m_kernel);
        OpenCL.clReleaseMemObject(m_clTextureDst);
        OpenCL.clReleaseMemObject(m_filterMask);
        OpenCL.clReleaseMemObject(m_clTextureSrc);
        OpenCL.clReleaseProgram(m_clProgram);
        OpenCL.clReleaseCommandQueue(m_queue);
        OpenCL.clReleaseContext(m_context);
        CLUtil.destroyCL();
        m_glTextureSrc.delete();
    }
    
    public void setMask()
    {
        if(m_filterMask != null)
        {
            OpenCL.clReleaseMemObject(m_filterMask);
        }
        
        float[] mask = getGaussianBlur(Settings.m_filterSize, Settings.m_sigma);
        
        float sum = 0;
        
        for(int i = 0; i < Settings.m_filterSize * Settings.m_filterSize; ++i)
        {
            sum += mask[i];
        }
        FloatBuffer maskBuffer = BufferUtils.createFloatBuffer(mask.length);
        maskBuffer.put(mask);
        maskBuffer.position(0);
        
        m_filterMask = OpenCL.clCreateBuffer(m_context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, maskBuffer);
        
        OpenCL.clSetKernelArg(m_kernel, 2, m_filterMask);
        OpenCL.clSetKernelArg(m_kernel, 3, Settings.m_filterSize);
        OpenCL.clSetKernelArg(m_kernel, 5, sum);
        
        Settings.print();
    }

    @Override
    public void increaseMaskSize() {
        Settings.m_filterSize += Settings.m_filterSizeDelta;
        Settings.m_filterSize = Math.min(Math.max(Settings.m_filterSize, 1), Settings.MAX_FILTER_SIZE);
        setMask();
    }

    @Override
    public void decreaseMaskSize() { 
        Settings.m_filterSize -= Settings.m_filterSizeDelta;
        Settings.m_filterSize = Math.min(Math.max(Settings.m_filterSize, 1), Settings.MAX_FILTER_SIZE);
        setMask();
    }

    @Override
    public void increaseStandardDevation() {
        Settings.m_sigma += Settings.m_sigmaDelta;
        setMask();
    }

    @Override
    public void decreaseStandardDevation() {
        Settings.m_sigma -= Settings.m_sigmaDelta;
        Settings.m_sigma = Math.max(Settings.m_sigma, 0.0001);
        setMask();
    }
}
